
#include "GeomIntr2DCirCir.h"
using namespace Geom;

//----------------------------------------------------------------------------
static bool Find (const Vector2& _C0, Real fR0, const Vector2& _C1,
    Real fR1, int& riQuantity, Vector2 akPoint[2])
{
    // The two circles are |X-C0| = R0 and |X-C1| = R1.  Define U = C1 - C0
    // and V = Perp(U) where Perp(x,y) = (y,-x).  Note that Dot(U,V) = 0 and
    // |V|^2 = |U|^2.  The intersection points X can be written in the form
    // X = C0+s*U+t*V and X = C1+(s-1)*U+t*V.  Squaring the circle equations
    // and substituting these formulas into them yields
    //   R0^2 = (s^2 + t^2)*|U|^2
    //   R1^2 = ((s-1)^2 + t^2)*|U|^2.
    // Subtracting and solving for s yields
    //   s = ((R0^2-R1^2)/|U|^2 + 1)/2
    // Then replace in the first equation and solve for t^2
    //   t^2 = (R0^2/|U|^2) - s^2.
    // In order for there to be solutions, the right-hand side must be
    // nonnegative.  Some algebra leads to the condition for existence of
    // solutions,
    //   (|U|^2 - (R0+R1)^2)*(|U|^2 - (R0-R1)^2) <= 0.
    // This reduces to
    //   |R0-R1| <= |U| <= |R0+R1|.
    // If |U| = |R0-R1|, then the circles are side-by-side and just tangent.
    // If |U| = |R0+R1|, then the circles are nested and just tangent.
    // If |R0-R1| < |U| < |R0+R1|, then the two circles to intersect in two
    // points.

    Vector2 kU = _C1 - _C0;
    Real fUSqrLen = kU.SquaredLength();
    if ( fUSqrLen == 0.0f && fR0 == fR1 )
    {
        // Circles are the same.  TO DO.  Allow an 'epsilon' for the floating
        // point compares?
        riQuantity = -1;
        return true;
    }

    Real fR0pR1 = fR0 + fR1;
    Real fR0pR1Sqr = fR0pR1*fR0pR1;
    if ( fUSqrLen > fR0pR1Sqr )
    {
        riQuantity = 0;
        return false;
    }

    Real fR0mR1 = fR0 - fR1;
    Real fR0mR1Sqr = fR0mR1*fR0mR1;
    if ( fUSqrLen < fR0mR1Sqr )
    {
        riQuantity = 0;
        return false;
    }

    if ( fUSqrLen < fR0pR1Sqr )
    {
        if ( fR0mR1Sqr < fUSqrLen )
        {
            Real fInvUSqrLen = 1.0f/fUSqrLen;
            Real fS = 0.5f*((fR0*fR0-fR1*fR1)*fInvUSqrLen+1.0f);
            Vector2 kTmp = _C0 + fS*kU;

            Real fT = Math::Sqrt(fR0*fR0*fInvUSqrLen - fS*fS);
            Vector2 kV(kU.y,-kU.x);
            riQuantity = 2;
            akPoint[0] = kTmp - fT*kV;
            akPoint[1] = kTmp + fT*kV;
        }
        else
        {
            // |U| = |R0-R1|, circles are tangent
            riQuantity = 1;
            akPoint[0] = _C0 + (fR0/fR0mR1)*kU;
        }
    }
    else
    {
        // |U| = |R0+R1|, circles are tangent
        riQuantity = 1;
        akPoint[0] = _C0 + (fR0/fR0pR1)*kU;
    }

    return true;
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Circle2& _Circle0,
    const Circle2& _Circle1, int& riQuantity, Vector2 akPoint[2])
{
    return Find(_Circle0.Center(),_Circle0.Radius(),_Circle1.Center(),
        _Circle1.Radius(),riQuantity,akPoint);
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Circle2& _Circle, const Arc2& _Arc,
    int& riQuantity, Vector2 akPoint[2])
{
    Vector2 akIntrPoint[2];
    bool bIntersects = Find(_Circle.Center(),_Circle.Radius(),
        _Arc.Center(),_Arc.Radius(),riQuantity,akIntrPoint);

    if ( riQuantity == -1 )
    {
        // arc is contained by the circle
        return true;
    }

    if ( bIntersects )
    {
        // test if circle-circle intersection points are on the arc
        int iIntrQuantity = riQuantity;
        riQuantity = 0;
        for (int i = 0; i < iIntrQuantity; i++)
        {
            akPoint[riQuantity] = akIntrPoint[i];
            if ( _Arc.Contains(akPoint[riQuantity]) )
                riQuantity++;
        }
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Arc2& _Arc0, const Arc2& _Arc1,
    int& riQuantity, Vector2 akPoint[2])
{
    Vector2 akIntrPoint[2];
    bool bIntersects = Find(_Arc0.Center(),_Arc0.Radius(),_Arc1.Center(),
        _Arc1.Radius(),riQuantity,akIntrPoint);

    if ( riQuantity == -1 )
    {
        // Arcs are cocircular.  Determine if they overlap.
        if ( _Arc1.Contains(_Arc0.End0()) )
        {
            if ( _Arc1.Contains(_Arc0.End1()) )
            {
                riQuantity = -2;
                akPoint[0] = _Arc0.End0();
                akPoint[1] = _Arc0.End1();
            }
            else
            {
                akPoint[0] = _Arc0.End0();
                if ( _Arc0.End0() != _Arc1.End1() )
                {
                    riQuantity = -2;
                    akPoint[1] = _Arc1.End1();
                }
                else
                {
                    riQuantity = -1;
                }
            }
            return true;
        }

        if ( _Arc1.Contains(_Arc0.End1()) )
        {
            akPoint[0] = _Arc1.End0();
            if ( _Arc0.End1() != _Arc1.End0() )
            {
                riQuantity = -2;
                akPoint[1] = _Arc0.End1();
            }
            else
            {
                riQuantity = -1;
            }
            return true;
        }

        Arc2 kTransArc;
        kTransArc.Center() = _Arc0.Center();
        kTransArc.Radius() = _Arc0.Radius();
        kTransArc.End0() = _Arc0.End0();
        kTransArc.End1() = _Arc1.End0();
        if ( kTransArc.Contains(_Arc0.End1()) )
        {
            riQuantity = 0;
            return false;
        }
        else
        {
            riQuantity = -2;
            akPoint[0] = _Arc1.End0();
            akPoint[1] = _Arc1.End1();
            return true;
        }
    }

    if ( bIntersects )
    {
        // test if circle-circle intersection points are on the arcs
        int iIntrQuantity = riQuantity;
        riQuantity = 0;
        for (int i = 0; i < iIntrQuantity; i++)
        {
            akPoint[riQuantity] = akIntrPoint[i];
            if ( _Arc0.Contains(akPoint[riQuantity])
            &&   _Arc1.Contains(akPoint[riQuantity]) )
            {
                riQuantity++;
            }
        }
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------


