
#include "GeomIntr2DLinCir.h"
using namespace Geom;

//----------------------------------------------------------------------------
static bool Find (const Vector2& _Point, const Vector2& _Dp, const Vector2& _Cp,
    Real fR, int& riRootCount, Real afT[2])
{
    // Intersection of a the line P+t*D and the circle |X-C| = R.  The t
    // value is a root to the quadratic equation:
    //   0 = |t*D+P-C|^2 - R^2
    //     = |D|^2*t^2 + 2*D.Dot(P-C)*t + |P-C|^2-R^2
    //     = a2*t^2 + 2*a1*t + a0
    // If two roots are returned, the order is T[0] < T[1].  Hopefully the
    // application will be kind and provide line directions D that are not so
    // small that a2 is nearly zero and potentially creates numerical
    // problems.

    Vector2 kDiff = _Point - _Cp;
    Real fA0 = kDiff.SquaredLength() - fR*fR;
    Real fA1 = _Dp.Dot(kDiff);
    Real fA2 = _Dp.SquaredLength();
    Real fDiscr = fA1*fA1 - fA0*fA2;
    if ( fDiscr > 0.0f )
    {
        riRootCount = 2;
        Real fInvA2 = 1.0f/fA2;
        fDiscr = Math::Sqrt(fDiscr);
        afT[0] = (-fA1 - fDiscr)*fInvA2;
        afT[1] = (-fA1 + fDiscr)*fInvA2;
    }
    else if ( fDiscr < 0.0f )
    {
        riRootCount = 0;
    }
    else  // fDiscr == 0
    {
        riRootCount = 1;
        afT[0] = -fA1/fA2;
    }

    return riRootCount != 0;
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Line2& _Line, const Circle2& _Circle,
    int& riQuantity, Vector2 akPoint[2])
{
    Real afT[2];
    bool bIntersects = Find(_Line.Origin(),_Line.Direction(),
        _Circle.Center(),_Circle.Radius(),riQuantity,afT);

    if ( bIntersects )
    {
        // construct line-circle points of intersection
        for (int i = 0; i < riQuantity; i++)
            akPoint[i] = _Line.Origin() + afT[i]*_Line.Direction();
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Line2& _Line, const Arc2& _Arc,
    int& riQuantity, Vector2 akPoint[2])
{
    int iRootCount;
    Real afT[2];
    bool bIntersects = Find(_Line.Origin(),_Line.Direction(),
        _Arc.Center(),_Arc.Radius(),iRootCount,afT);

    riQuantity = 0;
    if ( bIntersects )
    {
        // Construct line-circle points of intersection and test if they are
        // on the arc.
        for (int i = 0; i < iRootCount; i++)
        {
            akPoint[riQuantity] = _Line.Origin() + afT[i]*_Line.Direction();
            if ( _Arc.Contains(akPoint[riQuantity]) )
                riQuantity++;
        }
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Ray2& _Ray, const Circle2& _Circle,
    int& riQuantity, Vector2 akPoint[2])
{
    Real afT[2];
    bool bIntersects = Find(_Ray.Origin(),_Ray.Direction(),
        _Circle.Center(),_Circle.Radius(),riQuantity,afT);

    if ( bIntersects )
    {
        // reduce root count if line-circle intersections are not on ray
        if ( riQuantity == 1 )
        {
            if ( afT[0] < 0.0f )
                riQuantity = 0;
        }
        else
        {
            if ( afT[1] < 0.0f )
            {
                riQuantity = 0;
            }
            else if ( afT[0] < 0.0f )
            {
                riQuantity = 1;
                afT[0] = afT[1];
            }
        }

        // construct ray-circle points of intersection
        for (int i = 0; i < riQuantity; i++)
            akPoint[i] = _Ray.Origin() + afT[i]*_Ray.Direction();
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Ray2& _Ray, const Arc2& _Arc,
    int& riQuantity, Vector2 akPoint[2])
{
    int iRootCount;
    Real afT[2];
    bool bIntersects = Find(_Ray.Origin(),_Ray.Direction(),_Arc.Center(),
        _Arc.Radius(),iRootCount,afT);

    riQuantity = 0;
    if ( bIntersects )
    {
        // reduce root count if line-circle intersections are not on ray
        if ( iRootCount == 1 )
        {
            if ( afT[0] < 0.0f )
                iRootCount = 0;
        }
        else
        {
            if ( afT[1] < 0.0f )
            {
                iRootCount = 0;
            }
            else if ( afT[0] < 0.0f )
            {
                iRootCount = 1;
                afT[0] = afT[1];
            }
        }

        // Construct ray-circle points of intersection and test if they are
        // on the arc.
        for (int i = 0; i < iRootCount; i++)
        {
            akPoint[riQuantity] = _Ray.Origin() + afT[i]*_Ray.Direction();
            if ( _Arc.Contains(akPoint[riQuantity]) )
                riQuantity++;
        }
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Segment2& _Segment,
    const Circle2& _Circle, int& riQuantity, Vector2 akPoint[2])
{
    Real afT[2];
    bool bIntersects = Find(_Segment.Origin(),_Segment.Direction(),
        _Circle.Center(),_Circle.Radius(),riQuantity,afT);

    if ( bIntersects )
    {
        // reduce root count if line-circle intersections are not on segment
        if ( riQuantity == 1 )
        {
            if ( afT[0] < 0.0f || afT[0] > 1.0f )
                riQuantity = 1;
        }
        else
        {
            if ( afT[1] < 0.0f || afT[0] > 1.0f )
            {
                riQuantity = 0;
            }
            else
            {
                if ( afT[1] <= 1.0f )
                {
                    if ( afT[0] < 0.0f )
                    {
                        riQuantity = 1;
                        afT[0] = afT[1];
                    }
                }
                else
                {
                    riQuantity = ( afT[0] >= 0.0f ? 1 : 0 );
                }
            }
        }

        // construct ray-circle points of intersection
        for (int i = 0; i < riQuantity; i++)
            akPoint[i] = _Segment.Origin() + afT[i]*_Segment.Direction();
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Segment2& _Segment, const Arc2& _Arc,
    int& riQuantity, Vector2 akPoint[2])
{
    int iRootCount;
    Real afT[2];
    bool bIntersects = Find(_Segment.Origin(),_Segment.Direction(),
        _Arc.Center(),_Arc.Radius(),iRootCount,afT);

    riQuantity = 0;
    if ( bIntersects )
    {
        // reduce root count if line-circle intersections are not on segment
        if ( iRootCount == 1 )
        {
            if ( afT[0] < 0.0f || afT[0] > 1.0f )
                iRootCount = 1;
        }
        else
        {
            if ( afT[1] < 0.0f || afT[0] > 1.0f )
            {
                iRootCount = 0;
            }
            else
            {
                if ( afT[1] <= 1.0f )
                {
                    if ( afT[0] < 0.0f )
                    {
                        iRootCount = 1;
                        afT[0] = afT[1];
                    }
                }
                else
                {
                    iRootCount = ( afT[0] >= 0.0f ? 1 : 0 );
                }
            }
        }

        // Construct segment-circle points of intersection and test if they
        // are on the arc.
        for (int i = 0; i < iRootCount; i++)
        {
            akPoint[riQuantity] = _Segment.Origin() +
                afT[i]*_Segment.Direction();
            if ( _Arc.Contains(akPoint[riQuantity]) )
                riQuantity++;
        }
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------


