
#include "GeomIntr2DElpElp.h"
#include "GeomPolynomial.h"
using namespace Geom;

//----------------------------------------------------------------------------
static void GetPoly (const QuadPoly2& _P0, const QuadPoly2& _P1,
    Polynomial& _Poly)
{
    // polynomial is constructed as a Bezout determinant
    Real fAB = _P0.m_fA*_P1.m_fB-_P1.m_fA*_P0.m_fB;
    Real fAC = _P0.m_fA*_P1.m_fC-_P1.m_fA*_P0.m_fC;
    Real fAD = _P0.m_fA*_P1.m_fD-_P1.m_fA*_P0.m_fD;
    Real fAE = _P0.m_fA*_P1.m_fE-_P1.m_fA*_P0.m_fE;
    Real fAF = _P0.m_fA*_P1.m_fF-_P1.m_fA*_P0.m_fF;
    Real fBC = _P0.m_fB*_P1.m_fC-_P1.m_fB*_P0.m_fC;
    Real fBE = _P0.m_fB*_P1.m_fE-_P1.m_fB*_P0.m_fE;
    Real fBF = _P0.m_fB*_P1.m_fF-_P1.m_fB*_P0.m_fF;
    Real fCD = _P0.m_fC*_P1.m_fD-_P1.m_fC*_P0.m_fD;
    Real fDE = _P0.m_fD*_P1.m_fE-_P1.m_fD*_P0.m_fE;
    Real fDF = _P0.m_fD*_P1.m_fF-_P1.m_fD*_P0.m_fF;
    Real fBFpDE = fBF+fDE;
    Real fBEmCD = fBE-fCD;

    Real afU[5] =
    {
        fAD*fDF-fAF*fAF,
        fAB*fDF+fAD*fBFpDE-2.0f*fAE*fAF,
        fAB*fBFpDE+fAD*fBEmCD-fAE*fAE-2.0f*fAC*fAF,
        fAB*fBEmCD+fAD*fBC-2.0f*fAC*fAE,
        fAB*fBC-fAC*fAC
    };

    // reduce degree if necessary
    const Real fEpsilon = 1e-06f;
    int iDegree = 4;
    while ( iDegree > 0 && Math::FAbs(afU[iDegree]) < fEpsilon )
        iDegree--;

    _Poly.SetDegree(iDegree);
    memcpy(&_Poly[0],afU,(iDegree+1)*sizeof(Real));
}
//----------------------------------------------------------------------------
bool QuadPoly2::IsEllipse ()
{
    return m_fB*m_fB < 4.0f*m_fA*m_fC;
}
//----------------------------------------------------------------------------
void Geom::FindIntersection (const QuadPoly2& _E0, const QuadPoly2& _E1,
    int& riQuantity, Vector2 akP[4])
{
    Polynomial kPoly;
    GetPoly(_E0,_E1,kPoly);

    int iYCount;
    Real afY[4];
    Polynomial::ZERO_TOLERANCE = 1e-04f;
    switch ( kPoly.GetDegree() )
    {
    case 4: kPoly.RootsDegree4(iYCount,afY); break;
    case 3: kPoly.RootsDegree3(iYCount,afY); break;
    case 2: kPoly.RootsDegree2(iYCount,afY); break;
    case 1:
        iYCount = 1;
        afY[0] = -kPoly[0]/kPoly[1];
        break;
    case 0:
        riQuantity = 0;
        return;
    }

    // Adjustment for quadratics to allow for relative error testing.
    Real fNorm0 = _E0.m_fA*_E0.m_fA + 2.0f*_E0.m_fB*_E0.m_fB +
        _E0.m_fC*_E0.m_fC;
    Real fNorm1 = _E1.m_fA*_E1.m_fA + 2.0f*_E1.m_fB*_E1.m_fB +
        _E1.m_fC*_E1.m_fC;

    riQuantity = 0;
    int iXCount;
    Real afX[2];
    for (int iY = 0; iY < iYCount; iY++)
    {
        Polynomial kAPoly(2);
        kAPoly[0] = _E0.m_fF+afY[iY]*(_E0.m_fE+afY[iY]*_E0.m_fC);
        kAPoly[1] = _E0.m_fD+afY[iY]*_E0.m_fB;
        kAPoly[2] = _E0.m_fA;
        kAPoly.GetAllRoots(iXCount,afX);
        for (int iX = 0; iX < iXCount; iX++)
        {
            const Real fEpsilon = 1e-03f;
            Real fTest = (_E0.m_fA*afX[iX] + _E0.m_fB*afY[iY] +
                _E0.m_fD)*afX[iX] + (_E0.m_fC*afY[iY] + _E0.m_fE)*afY[iY]+
                _E0.m_fF;
            if ( Math::FAbs(fTest) < fEpsilon*fNorm0 )
            {
                fTest = (_E1.m_fA*afX[iX] + _E1.m_fB*afY[iY] + _E1.m_fD)*
                    afX[iX] + (_E1.m_fC*afY[iY] + _E1.m_fE)*afY[iY] +
                    _E1.m_fF;
                if ( Math::FAbs(fTest) < fEpsilon*fNorm1 )
                {
                    akP[riQuantity].x = afX[iX];
                    akP[riQuantity].y = afY[iY];
                    riQuantity++;
                }
            }
        }
    }
}
//----------------------------------------------------------------------------
bool Geom::TestIntersection (const QuadPoly2& _E0, const QuadPoly2& _E1)
{
    Polynomial kPoly;
    GetPoly(_E0,_E1,kPoly);
    return kPoly.GetRootCount(-Math::MAX_REAL,Math::MAX_REAL) > 0;
}
//----------------------------------------------------------------------------


